;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

	.nolist
	.title "msp430 eForth 4.3"
	.cdecls C,LIST,"msp430g2553.h"  ; Include device header file

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; 7/7/2012 430eForth1.0, from eForth86.asm and 430uForth
; 7/4/2012 Move 430uForth2.1 from IAR to CCS 5.2
; 8/5/2014 Move 430eForth2.2 to CCS 6.0.  Fix linkage of OVER.
;	Software UART at 2400 baud.
; 8/10/2014 430eForth2.3 9600 baud, thanks to Dirk Bruehl and
;	Michael Kalus of www.4e4th.org
; 8/10/2014 430eForth2.4 Restore ERASE and WRITE
; 8/20/2014 430eForth2.5 Test Segment D
; 8/25/2014 430eForth2.6 Turnkey
; 8/26/2014 430eForth2.7 Optimize
; 9/16/2014 430eForth3.1 Tail recursion, APP!
; 10/11/2014 430eForth4.1 Direct thread, more optimization
; 10/23/2014 430eForth4.2 Direct thread, pack lists
; 11/12/2014 430eForth4.2 Direct thread, final
;
; Build for and verified on MSP430G2 LaunchPad from TI
; Assembled with Code Composer Sudio 6.0 IDE
; Internal DCO at 8 MHz
; Hardware UART at 9600 baud. TXD and RXD must be crossed.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Direct Thread Model of eForth

;;	CPU registers
tos	.equ	R4
stack	.equ	R5
ip	.equ	R6
temp0	.equ	R7
temp1	.equ	R8
temp2	.equ	R9
temp3	.equ	R10

;; Macros

pops	.macro	;DROP
	mov.w	@stack+,tos
	.endm

pushs	.macro	;DUP
	decd.w	stack
	mov.w	tos,0(stack)
	.endm;; Constants

$NEXT	.macro
	mov @ip+,pc	; fetch code address into PC
	.endm

$NEST	.macro
	.align	2
	call	#DOLST	; fetch code address into PC, W=PFA
	.endm

$CONST	.macro
	.align	2
	call	#DOCON	; fetch code address into PC, W=PFA
	.endm

;; Assembler constants

COMPO	.equ	040H	;lexicon compile only bit
IMEDD	.equ	080H	;lexicon immediate bit
MASKK	.equ	07F1FH	;lexicon bit mask
CELLL	.equ	2	;size of a cell
BASEE	.equ	10	;default radix
VOCSS	.equ	8	;depth of vocabulary stack
BKSPP	.equ	8	;backspace
LF	.equ	10	;line feed
CRR	.equ	13	;carriage return
ERR	.equ	27	;error escape
TIC	.equ	39	;tick
CALLL	.equ	012B0H	;NOP CALL opcodes

UPP	.equ	200H
DPP	.equ	220H
SPP	.equ	378H	;data stack
TIBB	.equ	380H	;terminal input buffer
RPP	.equ	3F8H	;return stacl
CODEE	.equ	0C000H	;code dictionary
COLDD	.equ	0FFFEH	;cold start vector
EM	.equ	0FFFFH	;top of memory

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
	.list
;;;;
;; Main entry points and COLD start data

	.text
	.global	main
main:
init:
	mov	#RPP,SP	; set up stack
	mov	#SPP,stack
	clr	tos
	mov.w   #WDTPW+WDTHOLD,&WDTCTL  ; Stop watchdog timer
	bis.b   #041h,&P1DIR	; P1.0/6 output
	jmp	COLD

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Device dependent I/O

;   ?KEY	( -- F | c T )
;	Return input character.
	.word	0
	.byte	4,"?KEY",0
QKEY:
	pushs
QKEY1:
	BIT.B   #UCA0RXIFG,&IFG2
	JZ	FALSE	;return false flag
	MOV.B   &UCA0RXBUF,tos	; read character into TOS
	pushs
	jmp	TRUE

;   KEY	( -- c )
;	Return input character.
	.word	QKEY-6
	.byte	3,"KEY"
KEY
	pushs
KEY1:
	BIT.B   #UCA0RXIFG,&IFG2
	JZ	KEY1
	MOV.B   &UCA0RXBUF,tos	; read character into TOS
	$NEXT

;   EMIT	( c -- )
;	Send character c to the output device.
	.word	KEY-4
	.byte	4,"EMIT",0
EMIT
EMIT1:
	BIT.B   #UCA0TXIFG,&IFG2
	JZ	EMIT1
	MOV.B   tos,&UCA0TXBUF
	pops
	$NEXT

;   !IO	( -- )
;	Initialize the serial I/O devices.
;	.word	EMIT-6
;	.byte	3,"!IO"
STOIO
; 8MHz
	mov.b   &CALBC1_8MHZ, &BCSCTL1   ; Set DCO
	mov.b   &CALDCO_8MHZ, &DCOCTL	; to 8 MHz.
	mov.b   #006h, &P1SEL	; Use P1.1/P1.2 for USCI_A0
	mov.b   #006h, &P1SEL2	; Use P1.1/P1.2 for USCI_A0
; Configure UART (Koch)
	bis.b   #UCSSEL_2,&UCA0CTL1	;db2 SMCLK
	mov.b   #65,&UCA0BR0	;db3 8MHz 9600 Insgesamt &833 = $341
	mov.b   #3,&UCA0BR1	;db4 8MHz 9600
	mov.b   #UCBRS_2,&UCA0MCTL	;db5 Modulation UCBRSx = 2
	bic.b   #UCSWRST,&UCA0CTL1	;db6 **Initialize USCI
	$NEXT	;called from COLD

;; The kernel

;   doLIT	( -- w )
;	Push an inline literal.
;	.word	STOIO-4
;	.byte	COMPO+5,"doLIT"
DOLIT:
	pushs
	mov	@ip+,tos
	$NEXT

;   doCON	( -- a )
;	Run time routine for CONSTANT, VARIABLE and CREATE.
;	.word	DOLIT-6
;	.byte	COMPO+5,"doCON"
DOCON:
	pushs
	pop	tos
	mov	@tos,tos
	$NEXT

;   doLIST	( -- )
;	Process colon list..
;	.word	EMIT-6
;	.byte	6,"doLIST",0
DOLST
	mov	ip,temp0	;exchange pointers
	pop	ip	;push return stack
	push	temp0	;restore the pointers
	$NEXT

;   EXIT	( -- )
;	Terminate a colon definition.
	.word	EMIT-6
	.byte	4,"EXIT",0
EXIT
	mov	@sp+,ip
	$NEXT

;   EXECUTE	( ca -- )
;	Execute the word at ca.
	.word	EXIT-6
	.byte	7,"EXECUTE"
EXECU
	mov	tos,temp0
	pops
	br	temp0

;   @EXECUTE	( a -- )
;	Execute vector stored in address a.
	.word	EXECU-8
	.byte	8,"@EXECUTE",0
ATEXE
	mov	@tos,temp0
	pops
	br	temp0

;   branch	( -- )
;	Branch to an inline address.
;	.word	ATEXE-10
;	.byte	COMPO+6,"branch",0
BRAN
	mov	@ip+,ip
	$NEXT

;   ?branch	( f -- )
;	Branch if flag is zero.
;	.word	BRAN-8
;	.byte	COMPO+7,"?branch"
QBRAN
	tst	tos
	pops
	jz	BRAN
	jmp	SKIP

;   next	( -- )
;	Run time code for the single index loop.
;	: next ( -- ) \ hilevel model
;	r> r> dup if 1 - >r @ >r exit then drop cell+ >r ;
;	.word	QBRAB-8
;	.byte	COMPO+4,"next",0
DONXT
	dec	0(sp)   ;decrement index
	jge	BRAN	;loop back
	incd.w	sp	;discard index
SKIP:
	incd.w	ip	;exit loop
	$NEXT

;   !	( w a -- )
;	Pop the data stack to memory.
	.word	ATEXE-10
	.byte	1,"!"
STORE
	mov.w	@stack+,0(tos)
	pops
	$NEXT

;   @	( a -- w )
;	Push memory location to the data stack.
	.word	STORE-2
	.byte	1,"@"
AT
	mov.w	@tos,tos
	$NEXT

;   C!	( c b -- )
;	Pop the data stack to byte memory.
	.word	AT-2
	.byte	2,"C!",0
CSTOR
	mov.b	@stack+,0(tos)
	inc	stack
	pops
	$NEXT

;   C@	( b -- c )
;	Push byte memory location to the data stack.
	.word	CSTOR-4
	.byte	2,"C@",0
CAT
	mov.b	@tos,tos
	$NEXT

;   RP!	( -- )
;	init return stack pointer.
;	.word	CAT-4
;	.byte	3,"RP!"
RPSTO:
	mov	#RPP,SP	;init return stack
	$NEXT

;   R>	( -- w )
;	Pop the return stack to the data stack.
	.word	CAT-4
	.byte	2,"R",3EH,0
RFROM
	pushs
	pop	tos
	$NEXT

;   R@	( -- w )
;	Copy top of return stack to the data stack.
	.word	RFROM-4
	.byte	2,"R@",0
RAT
	pushs
	mov	0(sp),tos
	$NEXT

;   >R	( w -- )
;	Push the data stack to the return stack.
	.word	RAT-4
	.byte	COMPO+2,">R",0
TOR
	push	tos
	pops
	$NEXT

;   SP!	( -- )
;	Init data stack pointer.
;	.word	SPAT-4
;	.byte	3,"SP!"
SPSTO:
	mov	#SPP,stack	;init parameter stack
	clr	tos
	$NEXT

;   DROP	( w -- )
;	Discard top stack item.
	.word	TOR-4
	.byte	4,"DROP",0
DROP:
	pops
	$NEXT

;   DUP	( w -- w w )
;	Duplicate the top stack item.
	.word	DROP-6
	.byte	3,"DUP"
DUPP:
	pushs
	$NEXT

;   SWAP	( w1 w2 -- w2 w1 )
;	Exchange top two stack items.
	.word	DUPP-4
	.byte	4,"SWAP",0
SWAP
	mov.w	tos,temp0
	mov.w	@stack,tos
	mov.w	temp0,0(stack)
	$NEXT

;   OVER	( w1 w2 -- w1 w2 w1 )
;	Copy second stack item to top.
	.word	SWAP-6
	.byte	4,"OVER",0
OVER:
	mov.w	@stack,temp0
	pushs
	mov.w	temp0,tos
	$NEXT

;   0<	( n -- t )
;	Return true if n is negative.
	.word	OVER-6
	.byte	2,"0",3CH,0
ZLESS:
	tst	tos
	jn	TRUE
FALSE:
	clr	tos
	$NEXT
TRUE:
	mov	#0x-1,tos
	$NEXT

;   AND	( w w -- w )
;	Bitwise AND.
	.word	ZLESS-4
	.byte	3,"AND"
ANDD
	and	@stack+,tos
	$NEXT

;   OR	( w w -- w )
;	Bitwise inclusive OR.
	.word	ANDD-4
	.byte	2,"OR",0
ORR
	bis	@stack+,tos
	$NEXT

;   XOR	( w w -- w )
;	Bitwise exclusive OR.
	.word	ORR-4
	.byte	3,"XOR"
XORR
	xor	@stack+,tos
	$NEXT

;   UM+	( w w -- w cy )
;	Add two numbers, return the sum and carry flag.
	.word	XORR-4
	.byte	3,"UM+"
UPLUS
	add	@stack,tos
	mov	tos,0(stack)
	clr	tos
	rlc	tos
	$NEXT

;; Common functions

;   ?DUP	( w -- w w | 0 )
;	Dup tos if its is not zero.
	.word	UPLUS-4
	.byte	4,"?DUP",0
QDUP
	tst	tos
	jnz	DUPP
	$NEXT

;   ROT	( w1 w2 w3 -- w2 w3 w1 )
;	Rot 3rd item to top.
	.word	QDUP-6
	.byte	3,"ROT"
ROT
	mov.w	0(stack),temp0
	mov.w	tos,0(stack)
	mov.w	2(stack),tos
	mov.w	temp0,2(stack)
	$NEXT

;   2DROP	( w w -- )
;	Discard two items on stack.
	.word	ROT-4
	.byte	5,"2DROP"
DDROP
	incd.w	stack
	pops
	$NEXT

;   2DUP	( w1 w2 -- w1 w2 w1 w2 )
;	Duplicate top two items.
	.word	DDROP-6
	.byte	4,"2DUP",0
DDUP
	mov.w	@stack,temp0
	pushs
	decd.w	stack
	mov.w	temp0,0(stack)
	$NEXT

;   +	( w w -- sum )
;	Add top two items.
	.word	DDUP-6
	.byte	1,"+"
PLUS
	add	@stack+,tos
	$NEXT

;   D+	( d d -- d )
;	Double addition, as an example using UM+.
;
	.word	PLUS-2
	.byte	2,"D+",0
DPLUS
	mov.w	@stack+,temp0
	mov.w	@stack+,temp1
	add.w	temp0,0(stack)
	addc	temp1,tos
	$NEXT

;   NOT	( w -- w )
;	One's complement of tos.
	.word	DPLUS-4
	.byte	3,"NOT"
INVER
	inv	tos
	$NEXT

;   NEGATE	( n -- -n )
;	Two's complement of tos.
	.word	INVER-4
	.byte	6,"NEGATE",0
NEGAT
	inv	tos
	inc	tos
	$NEXT

;   DNEGATE	( d -- -d )
;	Two's complement of top double.
	.word	NEGAT-8
	.byte	7,"DNEGATE"
DNEGA:
	inv	tos
	inv 0(stack)
	inc	0(stack)
	addc	#0,tos
	$NEXT

;   -	( n1 n2 -- n1-n2 )
;	Subtraction.
	.word	DNEGA-8
	.byte	1,"-"
SUBB
	sub	@stack+,tos
	jmp	NEGAT

;   ABS	( n -- n )
;	Return the absolute value of n.
	.word	SUBB-2
	.byte	3,"ABS"
ABSS
	tst.w	tos
	jn	NEGAT
	$NEXT

;   =	( w w -- t )
;	Return true if top two are equal.
	.word	ABSS-4
	.byte	1,3DH
EQUAL
	xor	@stack+,tos
	jnz	FALSE
	jmp	TRUE

;   U<	( u u -- t )
;	Unsigned compare of top two items.
	.word	EQUAL-2
	.byte	2,"U",3CH,0
ULESS
	mov	@stack+,temp0
	cmp	tos,temp0
	subc	tos,tos
	$NEXT

;   <	( n1 n2 -- t )
;	Signed compare of top two items.
	.word	ULESS-4
	.byte	1,3CH
LESS
	cmp	@stack+,tos
	jz	FALSE
	jge	TRUE
	jmp	FALSE

;   >	( n1 n2 -- t )
;	Signed compare of top two items.
	.word	LESS-2
	.byte	1,3EH
GREAT
	cmp	@stack+,tos
	jge	FALSE
	jmp	TRUE

;   MAX	( n n -- n )
;	Return the greater of two top stack items.
	.word	GREAT-2
	.byte	3,"MAX"
MAX:
	cmp	0(stack),tos
MAX1:
	jl	DROP
	incd.w	stack
	$NEXT

;   MIN	( n n -- n )
;	Return the smaller of top two stack items.
	.word	MAX-4
	.byte	3,"MIN"
MIN:
	cmp	tos,0(stack)
	jmp	MAX1

;; Divide

;   UM/MOD	( udl udh u -- ur uq )
;	Unsigned divide of a double by a single. Return mod and quotient.
	.word	MIN-4
	.byte	6,"UM/MOD",0
UMMOD:
	mov	tos,temp0
	pops
	mov	#17,temp1
UMMOD2:
	cmp	temp0,tos
	jnc	UMMOD3
	sub	temp0,tos
	setc
	jmp	UMMOD4
UMMOD3:
	clrc
UMMOD4:
	rlc	0(stack)
	rlc	tos
	dec	temp1
	jnz	UMMOD2
	rra	tos
	mov	tos,temp0
	mov	0(stack),tos
	mov	temp0,0(stack)
	$NEXT

;   M/MOD	( d n -- r q )
;	Signed floored divide of double by single. Return mod and quotient.
	.word	UMMOD-8
	.byte	5,"M/MOD"
MSMOD
	$NEST
	.word	DUPP,ZLESS,DUPP,TOR,QBRAN,MMOD1
	.word	NEGAT,TOR,DNEGA,RFROM
MMOD1:
	.word	TOR,DUPP,ZLESS,QBRAN,MMOD2
	.word	RAT,PLUS
MMOD2:
	.word	RFROM,UMMOD,RFROM,QBRAN,MMOD3
	.word	SWAP,NEGAT,SWAP
MMOD3:
	.word	EXIT

;   /MOD	( n n -- r q )
;	Signed divide. Return mod and quotient.
	.word	MSMOD-6
	.byte	4,"/MOD",0
SLMOD
	$NEST
	.word	OVER,ZLESS,SWAP,MSMOD,EXIT

;   MOD	( n n -- r )
;	Signed divide. Return mod only.
	.word	SLMOD-6
	.byte	3,"MOD"
MODD
	$NEST
	.word	SLMOD,DROP,EXIT

;   /	( n n -- q )
;	Signed divide. Return quotient only.
	.word	MODD-4
	.byte	1,"/"
SLASH
	$NEST
	.word	SLMOD,SWAP,DROP,EXIT

;; Multiply

;   UM*	( u u -- ud )
;	Unsigned multiply. Return double product.
	.word	SLASH-2
	.byte	3,"UM*"
UMSTA:
	clr	temp0
	mov	#16,temp1
UMSTA2:
	bit	#1,0(stack)
	jz	UMSTA3
	add	tos,temp0
	jmp	UMSTA4
UMSTA3:
	clrc
UMSTA4:
	rrc	temp0
	rrc	0(stack)
	dec	temp1
	jnz	UMSTA2
	mov	temp0,tos
	$NEXT

;   *	( n n -- n )
;	Signed multiply. Return single product.
	.word	UMSTA-4
	.byte	1,"*"
STAR
	$NEST
	.word	UMSTA,DROP,EXIT

;   M*	( n n -- d )
;	Signed multiply. Return double product.
	.word	STAR-2
	.byte	2,"M*",0
MSTAR
	$NEST
	.word	DDUP,XORR,ZLESS,TOR
	.word	ABSS,SWAP,ABSS,UMSTA,RFROM
	.word	QBRAN,MSTA1
	.word	DNEGA
MSTA1:
	.word	EXIT

;   */MOD	( n1 n2 n3 -- r q )
;	Multiply n1 and n2, then divide by n3. Return mod and quotient.
	.word	MSTAR-4
	.byte	5,"*/MOD"
SSMOD
	$NEST
	.word	TOR,MSTAR,RFROM,MSMOD,EXIT

;   */	( n1 n2 n3 -- q )
;	Multiply n1 by n2, then divide by n3. Return quotient only.
	.word	SSMOD-6
	.byte	2,"*/",0
STASL
	$NEST
	.word	SSMOD,SWAP,DROP,EXIT

;; Miscellaneous

;   1+	( a -- a+1 )
;	Increment.
;	.word	STASL-4
;	.byte	2,"1+",0
ONEP
	add	#1,tos
	$NEXT

;   1-	( a -- a-1 )
;	Decrement
;	.word	ONEP-4
;	.byte	2,"1-",0
ONEM
	sub	#1,tos
	$NEXT

;   2+	( a -- a+2 )
;	Add cell size in byte to address.
;	.word	ONEM-4
;	.byte	2,"2+",0
CELLP
	add	#2,tos
	$NEXT

;   2-	( a -- a-2 )
;	Subtract cell size in byte from address.
;	.word	CELLP-4
;	.byte	2,"2-",0
CELLM
	sub	#2,tos
	$NEXT

;   2*	( n -- 2*n )
;	Multiply tos by cell size in bytes.
	.word	STASL-4
	.byte	2,"2*",0
CELLS
	rla	tos
	$NEXT

;   2/	( n -- n/2 )
;	Divide tos by cell size in bytes.
	.word	CELLS-4
	.byte	2,"2/",0
TWOSL:
	rra	tos
	$NEXT

;   ALIGNED	( b -- a )
;	Align address to the cell boundary.
;	.word	TWOSL-4
;	.byte	7,"ALIGNED"
ALGND
	add	#1,tos
	bic	#1,tos
	$NEXT

;   >CHAR	( c -- c )
;	Filter non-printing characters.
;	.word	TWOSL-4
;	.byte	5,">CHAR"
TCHAR
	$NEST
	.word	BLANK,MAX	;mask msb
	.word	DOLIT,126,MIN	;check for printable
	.word	EXIT

;   DEPTH	( -- n )
;	Return the depth of the data stack.
	.word	TWOSL-4
	.byte	5,"DEPTH"
DEPTH
	mov	stack,temp0
	pushs
	mov	#SPP,tos
	sub	temp0,tos
	rra	tos
	$NEXT

;   PICK	( ... +n -- ... w )
;	Copy the nth stack item to tos.
	.word	DEPTH-6
	.byte	4,"PICK",0
PICK
	rla	tos
	add stack,tos
	mov	@tos,tos
	$NEXT

;; Memory access

;   +!	( n a -- )
;	Add n to the contents at address a.
	.word	PICK-6
	.byte	2,"+!",0
PSTOR
	add @stack+,0(tos)
	pops
	$NEXT

;   COUNT	( b -- b +n )
;	Return count byte of a string and add 1 to byte address.
	.word	PSTOR-4
	.byte	5,"COUNT"
COUNT
	mov.b	@tos+,temp0
	pushs
	mov	temp0,tos
	$NEXT

;   CMOVE	( b1 b2 u -- )
;	Copy u bytes from b1 to b2.
	.word	COUNT-6
	.byte	5,"CMOVE"
CMOVE:
	mov	@stack+,temp0	;destination
	mov	@stack+,temp1	;source
	jmp	CMOVE2
CMOVE1:
	mov.b	@temp1+,0(temp0)
	inc	temp0
CMOVE2:
	dec	tos
	jn	CMOVE3	;I need a jp.  Oh, well.
	jmp	CMOVE1
CMOVE3:
	JMP	DROP

;   FILL	( b u c -- )
;	Fill u bytes of character c to area beginning at b.
	.word	CMOVE-6
	.byte	4,"FILL",0
FILL:
	mov	@stack+,temp0	;count
	mov	@stack+,temp1	;destination
	jmp	FIL2
FIL1:
	mov.b	tos,0(temp1)
	inc	temp1
FIL2:
	dec	temp0
	jn	FIL3
	jmp	FIL1
FIL3:
	JMP	DROP

;; User variables

;   'BOOT	( -- a )
;	The application startup vector.
	.word	FILL-6
	.byte	5,"'BOOT"
TBOOT
	$CONST
	.word	200H

;   BASE	( -- a )
;	Storage of the radix base for numeric I/O.
	.word	TBOOT-6
	.byte	4,"BASE",0
BASE
	$CONST
	.word	202H

;   tmp	( -- a )
;	A temporary storage location used in parse and find.
;	.word	BASE-6
;	.byte	COMPO+3,"tmp"
TEMP
	$CONST
	.word	204H

;   #TIB	( -- a )
;	Hold the character pointer while parsing input stream.
;	.word	BASE-6
;	.byte	4,"#TIB",0
NTIB
	$CONST
	.word	206H

;   >IN	( -- a )
;	Hold the character pointer while parsing input stream.
;	.word	NTIB-6
;	.byte	3,">IN"
INN
	$CONST
	.word	208H

;   HLD	( -- a )
;	Hold a pointer in building a numeric output string.
;	.word	INN-4
;	.byte	3,"HLD"
HLD
	$CONST
	.word	20AH

;   'EVAL	( -- a )
;	A area to specify vocabulary search order.
;	.word	HLD-4
;	.byte	5,"'EVAL"
TEVAL
	$CONST
	.word	20CH

;   CONTEXT	( -- a )
;	A area to specify vocabulary search order.
;	.word	TEVAL-6
;	.byte	7,"CONTEXT"
CNTXT
	$CONST
	.word	20EH

;   CP	( -- a )
;	Point to the top of the code dictionary.
	.word	BASE-6
	.byte	2,"CP",0
CP
	$CONST
	.word	210H

;   DP	( -- a )
;	Point to the bottom of the free ram area.
	.word	CP-4
	.byte	2,"DP",0
DP
	$CONST
	.word	212H

;   LAST	( -- a )
;	Point to the last name in the name dictionary.
;	.word	DP-4
;	.byte	4,"LAST",0
LAST
	$CONST
	.word	214H

;   HERE	( -- a )
;	Return the top of the code dictionary.
	.word	DP-4
	.byte	4,"HERE",0
HERE
	$NEST
	.word	DP,AT,EXIT

;   PAD	( -- a )
;	Return the address of a temporary buffer.
	.word	HERE-6
	.byte	3,"PAD"
PAD
	$NEST
	.word	HERE,DOLIT,80,PLUS,EXIT

;   TIB	( -- a )
;	Return the address of the terminal input buffer.
	.word	PAD-4
	.byte	3,"TIB"
TIB
	$CONST
	.word	TIBB

;; Numeric output, single precision

;   DIGIT	( u -- c )
;	Convert digit u to a character.
;	.word	LAST-6
;	.byte	5,"DIGIT"
DIGIT:
	cmp	#10,tos
	jl	DIGIT1
	add	#7,tos
DIGIT1:
	add	#"0",tos
	$NEXT

;   EXTRACT	( n base -- n c )
;	Extract the least significant digit from n.
;	.word	DIGIT-6
;	.byte	7,"EXTRACT"
EXTRC
	$NEST
	.word	DOLIT,0,SWAP,UMMOD
	.word	SWAP,DIGIT,EXIT

;   <#	( -- )
;	Initiate the numeric output process.
	.word	TIB-4
	.byte	2,"<#",0
BDIGS
	$NEST
	.word	PAD,HLD,STORE,EXIT

;   HOLD	( c -- )
;	Insert a character into the numeric output string.
	.word	BDIGS-4
	.byte	4,"HOLD",0
HOLD
	$NEST
	.word	HLD,AT,ONEM
	.word	DUPP,HLD,STORE,CSTOR,EXIT

;   #	( u -- u )
;	Extract one digit from u and append the digit to output string.
	.word	HOLD-6
	.byte	1,"#"
DIG
	$NEST
	.word	BASE,AT,EXTRC,HOLD,EXIT

;   #S	( u -- 0 )
;	Convert u until all digits are added to the output string.
	.word	DIG-2
	.byte	2,"#S",0
DIGS
	$NEST
DIGS1:
	.word	DIG,DUPP,QBRAN,DIGS2
	.word	BRAN,DIGS1
DIGS2:
	.word	EXIT

;   SIGN	( n -- )
;	Add a minus sign to the numeric output string.
	.word	DIGS-4
	.byte	4,"SIGN",0
SIGN
	$NEST
	.word	ZLESS,QBRAN,SIGN1
	.word	DOLIT,"-",HOLD
SIGN1:
	.word	EXIT

;   #>	( w -- b u )
;	Prepare the output string to be TYPE'd.
	.word	SIGN-6
	.byte	2,"#",3EH,0
EDIGS
	$NEST
	.word	DROP,HLD,AT
	.word	PAD,OVER,SUBB,EXIT

;   str	( n -- b u )
;	Convert a signed integer to a numeric string.
;	.word	EDIGS-4
;	.byte	3,"str"
STR
	$NEST
	.word	DUPP,TOR,ABSS
	.word	BDIGS,DIGS,RFROM
	.word	SIGN,EDIGS,EXIT

;   HEX	( -- )
;	Use radix 16 as base for numeric conversions.
	.word	EDIGS-4
	.byte	3,"HEX"
HEX:
	mov	#16,&0x202
	$NEXT

;   DECIMAL	( -- )
;	Use radix 10 as base for numeric conversions.
	.word	HEX-4
	.byte	7,"DECIMAL"
DECIM
	mov	#10,&0x202
	$NEXT

;; Numeric input, single precision

;   DIGIT?	( c base -- u t )
;	Convert a character to its numeric value. A flag indicates success.
;	.word	DECIM-8
;	.byte	6,"DIGIT?",0
DIGTQ
	mov	@stack,temp0
	sub	#"0",temp0
	jl	FALSE1
	cmp	#10,temp0
	jl	DIGTQ1
	sub	#7,temp0
DIGTQ1:
	cmp	tos,temp0
	mov	temp0,0(stack)
	jl	TRUE1
FALSE1:
	clr	tos
	$NEXT
TRUE1:
	mov	#-1,tos
	$NEXT

;   NUMBER?	( a -- n T | a F )
;	Convert a number string to integer. Push a flag on tos.
	.word	DECIM-8
	.byte	7,"NUMBER?"
NUMBQ
	$NEST
	.word	BASE,AT,TOR,DOLIT,0,OVER,COUNT
	.word	OVER,CAT,DOLIT,'$',EQUAL,QBRAN,NUMQ1
	.word	HEX,SWAP,ONEP,SWAP,ONEM
NUMQ1:
	.word	OVER,CAT,DOLIT,'-',EQUAL,TOR
	.word	SWAP,RAT,SUBB,SWAP,RAT,PLUS,QDUP
	.word	QBRAN,NUMQ6
	.word	ONEM,TOR
NUMQ2:
	.word	DUPP,TOR,CAT,BASE,AT,DIGTQ
	.word	QBRAN,NUMQ4
	.word	SWAP,BASE,AT,STAR,PLUS,RFROM,ONEP
	.word	DONXT,NUMQ2
	.word	RAT,SWAP,DROP,QBRAN,NUMQ3
	.word	NEGAT
NUMQ3:
	.word	SWAP
	.word	BRAN,NUMQ5
NUMQ4:
	.word	RFROM,RFROM,DDROP,DDROP,DOLIT,0
NUMQ5:
	.word	DUPP
NUMQ6:
	.word	RFROM,DDROP,RFROM,BASE,STORE,EXIT

;; Terminal output

;   BL	( -- 32 )
;	Return 32, the blank character.
	.word	NUMBQ-8
	.byte	2,"BL",0
BLANK
	$CONST
	.word	20H

;   SPACE	( -- )
;	Send the blank character to the output device.
	.word	BLANK-4
	.byte	5,"SPACE"
SPACE
	$NEST
	.word	BLANK,EMIT,EXIT

;   SPACES	( +n -- )
;	Send n spaces to the output device.
	.word	SPACE-6
	.byte	6,"SPACES",0
SPACS
	$NEST
	.word	DOLIT,0,MAX,TOR,BRAN,CHAR2
CHAR1:
	.word	SPACE
CHAR2:
	.word	DONXT,CHAR1,EXIT

;   TYPE	( b u -- )
;	Output u characters from b.
	.word	SPACS-8
	.byte	4,"TYPE",0
TYPEE
	$NEST
	.word	TOR,BRAN,TYPE2
TYPE1:
	.word	DUPP,CAT,TCHAR,EMIT
	.word	ONEP
TYPE2:
	.word	DONXT,TYPE1
	.word	DROP,EXIT

;   CR	( -- )
;	Output a carriage return and a line feed.
	.word	TYPEE-6
	.byte	2,"CR",0
CR
	$NEST
	.word	DOLIT,CRR,EMIT
	.word	DOLIT,LF,EMIT,EXIT

;   do$	( -- a )
;	Return the address of a compiled string.
;	.word	CR-4
;	.byte	COMPO+3,"do$"
DOSTR
	$NEST
	.word	RFROM,RAT,RFROM,COUNT,PLUS
	.word	ALGND,TOR,SWAP,TOR,EXIT

;   $"|	( -- a )
;	Run time routine compiled by $". Return address of a compiled string.
;	.word	CR-4
;	.byte	COMPO+3,"$""|"
STRQP
	$NEST
	.word	DOSTR,EXIT	;force a call to do$

;   ."|	( -- )
;	Run time routine of ." . Output a compiled string.
;	.word	STRQP-4
;	.byte	COMPO+3,".""|"
DOTQP
	$NEST
	.word	DOSTR,COUNT,TYPEE,EXIT

;   .R	( n +n -- )
;	Display an integer in a field of n columns, right justified.
	.word	CR-4
	.byte	2,".R",0
DOTR
	$NEST
	.word	TOR,STR,RFROM,OVER,SUBB
	.word	SPACS,TYPEE,EXIT

;   U.R	( u +n -- )
;	Display an unsigned integer in n column, right justified.
	.word	DOTR-4
	.byte	3,"U.R"
UDOTR
	$NEST
	.word	TOR,BDIGS,DIGS,EDIGS
	.word	RFROM,OVER,SUBB,SPACS,TYPEE,EXIT

;   U.	( u -- )
;	Display an unsigned integer in free format.
	.word	UDOTR-4
	.byte	2,"U.",0
UDOT
	$NEST
	.word	BDIGS,DIGS,EDIGS,SPACE,TYPEE,EXIT

;   .	( w -- )
;	Display an integer in free format, preceeded by a space.
	.word	UDOT-4
	.byte	1,"."
DOT
	$NEST
	.word	BASE,AT,DOLIT,10,XORR	;?decimal
	.word	QBRAN,DOT1
	.word	UDOT,EXIT	;no, display unsigned
DOT1:	.word	STR,SPACE,TYPEE,EXIT	;yes, display signed

;   ?	( a -- )
;	Display the contents in a memory cell.
	.word	DOT-2
	.byte	1,"?"
QUEST
	$NEST
	.word	AT,DOT,EXIT

;; Parsing

;   parse	( b u c -- b u delta ; <string> )
;	Scan string delimited by c. Return found string and its offset.
;	.word	QUEST-2
;	.byte	5,"parse"
PARS
	$NEST
	.word	TEMP,STORE,OVER,TOR,DUPP,QBRAN,PARS8
	.word	ONEM,TEMP,AT,BLANK,EQUAL,QBRAN,PARS3
	.word	TOR
PARS1:
	.word	BLANK,OVER,CAT	;skip leading blanks ONLY
	.word	SUBB,ZLESS,INVER,QBRAN,PARS2
	.word	ONEP,DONXT,PARS1
	.word	RFROM,DROP,DOLIT,0,DUPP,EXIT
PARS2:
	.word	RFROM
PARS3:
	.word	OVER,SWAP,TOR
PARS4:
	.word	TEMP,AT,OVER,CAT,SUBB	;scan for delimiter
	.word	TEMP,AT,BLANK,EQUAL,QBRAN,PARS5
	.word	ZLESS
PARS5:
	.word	QBRAN,PARS6
	.word	ONEP,DONXT,PARS4
	.word	DUPP,TOR,BRAN,PARS7
PARS6:
	.word	RFROM,DROP,DUPP,ONEP,TOR
PARS7:
	.word	OVER,SUBB,RFROM,RFROM,SUBB,EXIT
PARS8:
	.word	OVER,RFROM,SUBB,EXIT

;   PARSE	( c -- b u ; <string> )
;	Scan input stream and return counted string delimited by c.
;	.word	QUEST-2
;	.byte	5,"PARSE"
PARSE
	$NEST
	.word	TOR,TIB,INN,AT,PLUS	;current input buffer pointer
	.word	NTIB,AT,INN,AT,SUBB	;remaining count
	.word	RFROM,PARS,INN,PSTOR,EXIT

;   .(	( -- )
;	Output following string up to next ) .
	.word	QUEST-2
	.byte	IMEDD+2,".(",0
DOTPR
	$NEST
	.word	DOLIT,")",PARSE,TYPEE,EXIT

;   (	( -- )
;	Ignore following string up to next ) . A comment.
	.word	DOTPR-4
	.byte	IMEDD+1,"("
PAREN
	$NEST
	.word	DOLIT,")",PARSE,DDROP,EXIT

;   \	( -- )
;	Ignore following text till the end of line.
	.word	PAREN-2
	.byte	IMEDD+1,"\"
BKSLA
	$NEST
	.word	NTIB,AT,INN,STORE,EXIT

;   CHAR	( -- c )
;	Parse next word and return its first character.
	.word	BKSLA-2
	.byte	4,"CHAR",0
CHAR
	$NEST
	.word	BLANK,PARSE,DROP,CAT,EXIT

;   TOKEN	( -- a ; <string> )
;	Parse a word from input stream and copy it to name dictionary.
	.word	CHAR-6
	.byte	5,"TOKEN"
TOKEN
	$NEST
	.word	BLANK,PARSE,DOLIT,31,MIN
TOKEN1
	.word	HERE,DDUP,CSTOR,ONEP
	.word	SWAP,CMOVE,HERE
	.word	DOLIT,0,HERE,COUNT,PLUS,CSTOR,EXIT

;   WORD	( c -- a ; <string> )
;	Parse a word from input stream and copy it to code dictionary.
	.word	TOKEN-6
	.byte	4,"WORD",0
WORDD
	$NEST
	.word	PARSE,BRAN,TOKEN1

;; Dictionary search

;   NAME>	( na -- ca )
;	Return a code address given a name address.
;	.word	WORDD-6
;	.byte	5,"NAME>"
NAMET
	mov.b	@tos+,temp0
	and	#0x1F,temp0
	add	temp0,tos
	inc	tos
	bic	#1,tos
	$NEXT

;   SAME?	( a a -- a a f \ -0+ )
;	Compare u cells in two strings. Return 0 if identical.
;	.word	NAMET-6
;	.byte	5,"SAME?"
SAMEQ
	pushs
	mov	2(stack),tos
	mov.b	@tos,tos
SAME1:
	mov	2(stack),temp0
	add	tos,temp0
	mov.b	0(temp0),temp0
	mov	0(stack),temp1
	add	tos,temp1
	mov.b	0(temp1),temp1
	sub	temp1,temp0
	jnz	SAME2
	dec	tos
	jnz	SAME1
	$NEXT
SAME2:
	jmp TRUE1

;   NAME?	( a -- ca na | a F )
;	Search all context vocabularies for a string.
;	.word	WORDD-6
;	.byte	5,"NAME?"
NAMEQ
	$NEST
	.word	CNTXT,AT
FIND1:
	.word	DUPP,QBRAN,FIND3	;end of dictionary
	.word	OVER,AT,OVER,AT,DOLIT,MASKK,ANDD,EQUAL
	.word	QBRAN,FIND4
	.word	SAMEQ,QBRAN,FIND2	;match
FIND4:
	.word	CELLM,AT,BRAN,FIND1
FIND2:
	.word	SWAP,DROP,DUPP,NAMET,SWAP,EXIT
FIND3:
	.word	EXIT

;; Terminal input

;   ^H	( bot eot cur -- bot eot cur )
;	Backup the cursor by one character.
;	.word	NAMEQ-6
;	.byte	2,"^H",0
BKSP
	$NEST
	.word	TOR,OVER,RFROM,SWAP,OVER,XORR
	.word	QBRAN,BACK1
	.word	DOLIT,BKSPP,EMIT,ONEM
	.word	BLANK,EMIT,DOLIT,BKSPP,EMIT
BACK1:
	.word	EXIT

;   TAP	( bot eot cur c -- bot eot cur )
;	Accept and echo the key stroke and bump the cursor.
;	.word	BKSP-4
;	.byte	3,"TAP"
TAP
	$NEST
	.word	DUPP,EMIT,OVER,CSTOR,ONEP,EXIT

;   kTAP	( bot eot cur c -- bot eot cur )
;	Process a key stroke, CR or backspace.
;	.word	TAP-4
;	.byte	4,"kTAP",0
KTAP
	$NEST
	.word	DUPP,DOLIT,CRR,XORR,QBRAN,KTAP2
	.word	DOLIT,BKSPP,XORR,QBRAN,KTAP1
	.word	BLANK,TAP,EXIT
KTAP1:
	.word	BKSP,EXIT
KTAP2:
	.word	DROP,SWAP,DROP,DUPP,EXIT

;   accept	( b u -- b u )
;	Accept characters to input buffer. Return with actual count.
	.word	WORDD-6
	.byte	6,"ACCEPT",0
ACCEP
	$NEST
	.word	OVER,PLUS,OVER
ACCP1:
	.word	DDUP,XORR,QBRAN,ACCP4
	.word	KEY,DUPP,BLANK,SUBB,DOLIT,95,ULESS
	.word	QBRAN,ACCP2
	.word	TAP,BRAN,ACCP1
ACCP2:
	.word	KTAP
ACCP3:
	.word	BRAN,ACCP1
ACCP4:
	.word	DROP,OVER,SUBB,EXIT

;   QUERY	( -- )
;	Accept input stream to terminal input buffer.
	.word	ACCEP-8
	.byte	5,"QUERY"
QUERY
	$NEST
	.word	TIB,DOLIT,80,ACCEP,NTIB,STORE
	.word	DROP,DOLIT,0,INN,STORE,EXIT

;; Error handling

; QUIT inits return stack. ERROR inits both stacks.

;   ERROR	( a -- )
;	Return address of a null string with zero count.
;	.word	QUERY-6
;	.byte	5,"ERROR"
ERROR:
	$NEST
	.word	SPACE,COUNT,TYPEE,DOLIT
	.word	3FH,EMIT,CR,SPSTO,QUIT

;   abort"	( f -- )
;	Run time routine of ABORT" . Abort with a message.
;	.word	ERROR-6
;	.byte	COMPO+6,"abort""",0
ABORQ
	$NEST
	.word	QBRAN,ABOR1	;text flag
	.word	DOSTR,COUNT,TYPEE,SPSTO,QUIT	;pass error string
ABOR1:
	.word	DOSTR,DROP,EXIT

;; Text interpreter

;   $INTERPRET	( a -- )
;	Interpret a word. If failed, try to convert it to an integer.
;	.word	ERROR-6
;	.byte	10,"$INTERPRET",0
INTER
	$NEST
	.word	NAMEQ,QDUP	;?defined
	.word	QBRAN,INTE1
	.word	AT,DOLIT,COMPO,ANDD	;?compile only lexicon bits
	.word	ABORQ
	.byte	13," compile only"
	.word	EXECU,EXIT	;execute defined word
INTE1:
	.word	NUMBQ	;convert a number
	.word	QBRAN,INTE2,EXIT
INTE2:
	.word	ERROR	;error

;   [	( -- )
;	Start the text interpreter.
	.word	QUERY-6
	.byte	IMEDD+1,"["
LBRAC
	$NEST
	.word	DOLIT,INTER,TEVAL,STORE,EXIT

;   .OK	( -- )
;	Display 'ok' only while interpreting.
;	.word	LBRAC-2
;	.byte	3,".OK"
DOTOK
	$NEST
	.word	DOLIT,INTER,TEVAL,AT,EQUAL
	.word	QBRAN,DOTO1
	.word	DOTQP
	.byte	3," ok"
DOTO1:	.word	CR,EXIT

;   ?STACK	( -- )
;	Abort if the data stack underflows.
;	.word	DOTOK-4
;	.byte	6,"?STACK",0
QSTAC
	$NEST
	.word	DEPTH,ZLESS	;check only for underflow
	.word	ABORQ
	.byte	10," underflow",0
	.word	EXIT

;   EVAL	( -- )
;	Interpret the input stream.
;	.word	QSTAC-8
;	.byte	4,"EVAL",0
EVAL
	$NEST
EVAL1:
	.word	TOKEN,DUPP,CAT	;?input stream empty
	.word	QBRAN,EVAL2
	.word	TEVAL,ATEXE,QSTAC	;evaluate input, check stack
	.word	BRAN,EVAL1
EVAL2:
	.word	DROP,DOTOK,EXIT	;prompt

;   QUIT	( -- )
;	Reset return stack pointer and start text interpreter.
	.word	LBRAC-2
	.byte	4,"QUIT",0
QUIT
	$NEST
	.word	RPSTO,LBRAC	;start interpretation
QUIT1:
	.word	QUERY,EVAL	;get input
	.word	BRAN,QUIT1	;continue till error

;; Compiler utilities

;   ALLOT	( n -- )
;	Allocate n bytes to the RAM dictionary.
	.word	QUIT-6
	.byte	5,"ALLOT"
ALLOT
	$NEST
	.word	DP,PSTOR,EXIT	;adjust code pointer

;   IALLOT	( n -- )
;	Allocate n bytes to the code dictionary.
	.word	ALLOT-6
	.byte	6,"IALLOT",0
IALLOT
	$NEST
	.word	CP,PSTOR,EXIT	;adjust code pointer

;   I!	( n a -- )
;	Store n to address a in code dictionary.
	.word	IALLOT-8
	.byte	2,"I!",0
ISTORE
	mov	#FWKEY,&FCTL3 ; Clear LOCK
	mov	#FWKEY+WRT,&FCTL1 ; Enable write
;	call	#STORE
	mov.w	@stack+,0(tos)
	pops
	mov	#FWKEY,&FCTL1 ; Done. Clear WRT
	mov	#FWKEY+LOCK,&FCTL3 ; Set LOCK
	$NEXT

;   ERASE	( a -- )
;	Erase a segment at address a.
	.word	ISTORE-4
	.byte	5,"ERASE"
IERASE
	mov	#FWKEY,&FCTL3 ; Clear LOCK
	mov	#FWKEY+ERASE,&FCTL1 ; Enable erase
	clr	0(tos)
	mov	#FWKEY+LOCK,&FCTL3 ; Set LOCK
	pops
	$NEXT

;   WRITE	( src dest n -- )
;	Copy n byte from src to dest.  Dest is in flash memory.
	.word	IERASE-6
	.byte	5,"WRITE"
WRITE
	$NEST
	.word	TWOSL,TOR
WRITE1
	.word	OVER,AT,OVER,ISTORE
	.word	CELLP,SWAP,CELLP,SWAP
	.word	DONXT,WRITE1
	.word	DDROP,EXIT

;   ,	( w -- )
;	Compile an integer into the code dictionary.
	.word	WRITE-6
	.byte	1,","
COMMA
	$NEST
	.word	CP,AT,DUPP,CELLP	;cell boundary
	.word	CP,STORE,ISTORE,EXIT

;   call,	( w -- )
;	Compile a call instruction into the code dictionary.
;	.word	COMMA-2
;	.byte	5,"call,"
CALLC
	$NEST
	.word	DOLIT,CALLL,COMMA
	.word	COMMA,EXIT

;   [COMPILE]	( -- ; <string> )
;	Compile the next immediate word into code dictionary.
;	.word	COMMA-2
;	.byte	IMEDD+9,"[COMPILE]"
BCOMP
	$NEST
	.word	TICK,COMMA,EXIT

;   COMPILE	( -- )
;	Compile the next address in colon list to code dictionary.
;	.word	BCOMP-10
;	.byte	COMPO+7,"COMPILE"
COMPI
	$NEST
	.word	RFROM,DUPP,AT,COMMA	;compile address
	.word	CELLP,TOR,EXIT	;adjust return address

;   LITERAL	( w -- )
;	Compile tos to code dictionary as an integer literal.
;	.word	COMPI-8
;	.byte	IMEDD+7,"LITERAL"
LITER
	$NEST
	.word	COMPI,DOLIT,COMMA,EXIT

;   $,"	( -- )
;	Compile a literal string up to next " .
;	.word	LITER-8
;	.byte	3,"$,"""
STRCQ
	$NEST
	.word	DOLIT,""""
	.word	WORDD	;move string to code dictionary
	.word	STRCQ1,EXIT

STRCQ1
	$NEST
	.word	DUPP,CAT,TWOSL	;calculate aligned end of string
	.word	TOR
STRCQ2
	.word	DUPP,AT,COMMA,CELLP
	.word	DONXT,STRCQ2
	.word	DROP,EXIT

;; Structures

;   FOR	( -- a )
;	Start a FOR-NEXT loop structure in a colon definition.
	.word	COMMA-2
	.byte	IMEDD+3,"FOR"
FOR
	$NEST
	.word	COMPI,TOR,BEGIN,EXIT

;   BEGIN	( -- a )
;	Start an infinite or indefinite loop structure.
	.word	FOR-4
	.byte	IMEDD+5,"BEGIN"
BEGIN
	$NEST
	.word	CP,AT,EXIT

;   NEXT	( a -- )
;	Terminate a FOR-NEXT loop structure.
	.word	BEGIN-6
	.byte	IMEDD+4,"NEXT",0
NEXT
	$NEST
	.word	COMPI,DONXT,COMMA,EXIT

;   UNTIL	( a -- )
;	Terminate a BEGIN-UNTIL indefinite loop structure.
	.word	NEXT-6
	.byte	IMEDD+5,"UNTIL"
UNTIL
	$NEST
	.word	COMPI,QBRAN,COMMA,EXIT

;   AGAIN	( a -- )
;	Terminate a BEGIN-AGAIN infinite loop structure.
	.word	UNTIL-6
	.byte	IMEDD+5,"AGAIN"
AGAIN
	$NEST
	.word	COMPI,BRAN,COMMA,EXIT

;   IF	( -- A )
;	Begin a conditional branch structure.
	.word	AGAIN-6
	.byte	IMEDD+2,"IF",0
IFF
	$NEST
	.word	COMPI,QBRAN,BEGIN
	.word	DOLIT,2,IALLOT,EXIT

;   AHEAD	( -- A )
;	Compile a forward branch instruction.
;	.word	IFF-4
;	.byte	IMEDD+5,"AHEAD"
AHEAD
	$NEST
	.word	COMPI,BRAN,BEGIN
	.word	DOLIT,2,IALLOT,EXIT

;   REPEAT	( A a -- )
;	Terminate a BEGIN-WHILE-REPEAT indefinite loop.
	.word	IFF-4
	.byte	IMEDD+6,"REPEAT",0
REPEA
	$NEST
	.word	AGAIN,BEGIN,SWAP,ISTORE,EXIT

;   THEN	( A -- )
;	Terminate a conditional branch structure.
	.word	REPEA-8
	.byte	IMEDD+4,"THEN",0
THENN
	$NEST
	.word	BEGIN,SWAP,ISTORE,EXIT

;   AFT	( a -- a A )
;	Jump to THEN in a FOR-AFT-THEN-NEXT loop the first time through.
	.word	THENN-6
	.byte	IMEDD+3,"AFT"
AFT
	$NEST
	.word	DROP,AHEAD,BEGIN,SWAP,EXIT

;   ELSE	( A -- A )
;	Start the false clause in an IF-ELSE-THEN structure.
	.word	AFT-4
	.byte	IMEDD+4,"ELSE",0
ELSEE
	$NEST
	.word	AHEAD,SWAP,THENN,EXIT

;   WHILE	( a -- A a )
;	Conditional branch out of a BEGIN-WHILE-REPEAT loop.
	.word	ELSEE-6
	.byte	IMEDD+5,"WHILE"
WHILE
	$NEST
	.word	IFF,SWAP,EXIT

;   ABORT"	( -- ; <string> )
;	Conditional abort with an error message.
	.word	WHILE-6
	.byte	IMEDD+6,"ABORT""",0
ABRTQ
	$NEST
	.word	COMPI,ABORQ,STRCQ,EXIT

;   $"	( -- ; <string> )
;	Compile an inline string literal.
	.word	ABRTQ-8
	.byte	IMEDD+2,"$""",0
STRQ
	$NEST
	.word	COMPI,STRQP,STRCQ,EXIT

;   ."	( -- ; <string> )
;	Compile an inline string literal to be typed out at run time.
	.word	STRQ-4
	.byte	IMEDD+2,".""",0
DOTQ
	$NEST
	.word	COMPI,DOTQP,STRCQ,EXIT

;; Colon compiler

;   ?UNIQUE	( a -- a )
;	Display a warning message if the word already exists.
;	.word	DOTQ-4
;	.byte	7,"?UNIQUE"
UNIQU
	$NEST
	.word	DUPP,NAMEQ	;?name exists
	.word	QBRAN,UNIQ1	;redefinitions are OK
	.word	DOTQP
	.byte	7," reDef "	;but warn the user
	.word	OVER,COUNT,TYPEE	;just in case its not planned
UNIQ1:
	.word	DROP,EXIT

;   $,n	( na -- )
;	Build a new dictionary name using the string at na.
;	.word	UNIQU-8
;	.byte	3,"$,n"
SNAME
	$NEST
	.word	DUPP,CAT	;?null input
	.word	QBRAN,SNAM1
	.word	UNIQU	;?redefinition
	.word	LAST,AT,COMMA	;save na for vocabulary link
	.word	CP,AT,LAST,STORE
	.word	STRCQ1,EXIT	;fill name field
SNAM1
	.word	STRQP
	.byte	5," name"	;null input
	.word	ERROR

;   $COMPILE	( a -- )
;	Compile next word to code dictionary as a token or literal.
;	.word	UNIQU-8
;	.byte	8,"$COMPILE",0
SCOMP
	$NEST
	.word	NAMEQ,QDUP	;?defined
	.word	QBRAN,SCOM2
	.word	AT,DOLIT,IMEDD,ANDD	;?immediate
	.word	QBRAN,SCOM1
	.word	EXECU,EXIT	;its immediate, execute
SCOM1:	.word	COMMA,EXIT	;its not immediate, compile
SCOM2:	.word	NUMBQ	;try to convert to number
	.word	QBRAN,SCOM3
	.word	LITER,EXIT	;compile number as integer
SCOM3:	.word	ERROR	;error

;   OVERT	( -- )
;	Link a new word into the current vocabulary.
;	.word	SCOMP-10
;	.byte	5,"OVERT"
OVERT
	$NEST
	.word	LAST,AT,CNTXT,STORE,EXIT

;   ;	( -- )
;	Terminate a colon definition.
	.word	DOTQ-4
	.byte	IMEDD+COMPO+1,";"
SEMIS
	$NEST
	.word	COMPI,EXIT,LBRAC,OVERT,EXIT

;   ]	( -- )
;	Start compiling the words in the input stream.
	.word	SEMIS-2
	.byte	1,"]"
RBRAC
	$NEST
	.word	DOLIT,SCOMP,TEVAL,STORE,EXIT

;   :	( -- ; <string> )
;	Start a new colon definition using next word as its name.
	.word	RBRAC-2
	.byte	1,":"
COLON
	$NEST
	.word	TOKEN,SNAME,DOLIT,DOLST
	.word	CALLC,RBRAC,EXIT

;; Defining words

;   HEADER	( -- ; <string> )
;	Compile a new array entry without allocating code space.
;	.word	DOCON-6
;	.byte	6,"HEADER",0
HEADER
	$NEST
	.word	TOKEN,SNAME,OVERT
	.word	DOLIT,DOCON,CALLC,EXIT

;   CREATE	( -- ; <string> )
;	Compile a new array entry without allocating code space.
	.word	COLON-2
	.byte	6,"CREATE",0
CREAT
	$NEST
	.word	HEADER,DP,AT,COMMA,EXIT

;   CONSTANT	( n -- ; <string> )
;	Compile a new constant.
	.word	CREAT-8
	.byte	8,"CONSTANT",0
CONST
	$NEST
	.word	HEADER,COMMA,EXIT

;   VARIABLE	( -- ; <string> )
;	Compile a new variable initialized to 0.
	.word	CONST-10
	.byte	8,"VARIABLE",0
VARIA
	$NEST
	.word	CREAT,DOLIT,2,ALLOT,EXIT

;; Tools

;   '	( -- ca )
;	Search context vocabularies for the next word in input stream.
	.word	VARIA-10
	.byte	1,"'"
TICK
	$NEST
	.word	TOKEN,NAMEQ	;?defined
	.word	QBRAN,TICK1
	.word	EXIT	;yes, push code address
TICK1:
	.word	ERROR	;no, error

;   DUMP	( a u -- )
;	Dump u bytes from a, in a formatted manner.
	.word	TICK-2
	.byte	4,"DUMP",0
DUMP
	$NEST
	.word	DOLIT,7,TOR	;start count down loop
DUMP1:
	.word	CR,DUPP,DOLIT,5,UDOTR
	.word	DOLIT,15,TOR
DUMP2:
	.word	COUNT,DOLIT,3,UDOTR
	.word	DONXT,DUMP2	;loop till done
	.word	SPACE,DUPP,DOLIT,16,SUBB
	.word	DOLIT,16,TYPEE	;display printable characters
	.word	DONXT,DUMP1	;loop till done
	.word	DROP,EXIT

;   .S	( ... -- ... )
;	Display the contents of the data stack.
	.word	DUMP-6
	.byte	2,".S",0
DOTS
	$NEST
	.word	CR,DEPTH	;stack depth
	.word	TOR	;start count down loop
	.word	BRAN,DOTS2	;skip first pass
DOTS1:
	.word	RAT,PICK,DOT	;index stack, display contents
DOTS2:
	.word	DONXT,DOTS1	;loop till done
	.word	DOTQP
	.byte	4," <sp",0
	.word	EXIT

;   >NAME	( ca -- na | F )
;	Convert code address to a name address.
;	.word	DOTS-4
;	.byte	5,">NAME"
TNAME
	$NEST
	.word	TOR,CNTXT,AT	;vocabulary link
TNAM1:
	.word	DUPP,QBRAN,TNAM2
	.word	DUPP,NAMET,RAT,XORR	;compare
	.word	QBRAN,TNAM2
	.word	CELLM	;continue with next word
	.word	AT,BRAN,TNAM1
TNAM2:
	.word	RFROM,DROP,EXIT

;   .ID	( na -- )
;	Display the name at address.
;	.word	TNAME-6
;	.byte	3,".ID"
DOTID
	$NEST
	.word	COUNT,DOLIT,01FH,ANDD	;mask lexicon bits
	.word	TYPEE,EXIT

;   WORDS	( -- )
;	Display the names in the context vocabulary.
	.word	DOTS-4
	.byte	5,"WORDS"
WORDS
	$NEST
	.word	CR,CNTXT,AT	;only in context
WORS1:
	.word	QDUP	;?at end of list
	.word	QBRAN,WORS2
	.word	DUPP,SPACE,DOTID	;display a name
	.word	CELLM,AT,BRAN,WORS1
WORS2:
	.word	EXIT

;; Cold boot

;   HI	( -- )
;	Display the sign-on message of eForth.
	.word	WORDS-6
	.byte	2,"HI",0
HI
	$NEST
	.word	CR,DOTQP
	.byte	14,"430eForth v4.3",0	;model
	.word	CR,EXIT

;   APP!	( a -- )	Turnkey
;	HEX : APP! 200 ! 1000 ERASE 200 1000 20 WRITE ;
	.word	HI-4
	.byte	4,"APP!",0
APPST
	$NEST
	.word	TBOOT,STORE,DOLIT,0x1000,IERASE
	.word	TBOOT,DOLIT,0x1000,DOLIT,0x20
	.word	WRITE,EXIT

;   COLD	( -- )
;	The hilevel cold start sequence.
	.word	APPST-6
	.byte	4,"COLD",0
COLD
	$NEST
	.word	STOIO
	.word	DOLIT,UZERO,DOLIT,UPP
	.word	DOLIT,ULAST-UZERO,CMOVE	;initialize user area
	.word	TBOOT,ATEXE	;application boot
	.word	QUIT	;start interpretation

CTOP

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;
; COLD start moves the following to user variables.

	.sect	".infoD"
UZERO:
	.word	HI	;200H, boot routine
	.word	BASEE	;202H, BASE
	.word	0	;204H, tmp
	.word	0	;206H, >IN
	.word	0	;208H, #TIB
	.word	0	;20AH, HLD
	.word	INTER	;20CH, 'EVAL
	.word	COLD-6	;20EH, CONTEXT pointer
	.word	CTOP+8	;210H, CP; pass ISR
	.word	DPP	;220H, DP
	.word	COLD-6	;214H, LAST
ULAST:

;===============================================================
	.sect   ".reset"	; MSP430 RESET Vector
	.short  main	;
	.end
;===============================================================

